import pathlib
import shutil
import unittest


class MockMixin:
    @property
    def _base_path(self):
        return pathlib.Path('./__tmp__').resolve()


class TestCase(unittest.TestCase):
    CONFIG_NAME = 'config'

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(pathlib.Path('./__tmp__'), ignore_errors=True)

    def assertFileExists(self, filepath: pathlib.Path) -> None:
        if filepath.exists():
            return
        raise AssertionError('File %s does not exist.' % filepath)

    def make_config(self, klass, name=''):
        return klass(
            self.__class__.__name__,
            f'{name}_{self.CONFIG_NAME}',
            autosave=False,
        )



# flake8: noqa
from platform import system

import xdgconfig.mixins as mixins


__version__ = '0.4.1'


if system() == 'Windows':
    from xdgconfig.config_win import WinConfig as Config
elif system() in ('Darwin', 'Linux') or system().startswith('CYGWIN'):
    from xdgconfig.config_unix import UnixConfig as Config
else:
    raise ImportError(
        "xdgconfig is not available on this platform : %s" % system()
    )

from xdgconfig.config import LocalConfig


class JsonConfig(mixins.JsonMixin, Config):
    ...


class LocalJsonConfig(mixins.JsonMixin, LocalConfig):
    ...


class IniConfig(mixins.IniMixin, Config):
    ...


class LocalIniConfig(mixins.IniMixin, LocalConfig):
    ...


if hasattr(mixins, 'XmlMixin'):
    class XmlConfig(mixins.XmlMixin, Config):
        ...


    class LocalXmlConfig(mixins.XmlMixin, LocalConfig):
        ...
else:
    print((
        'xmltodict is not installed. '
        'Run pip install xdgconfig[xml] to install it.'
    ))

if hasattr(mixins, 'YamlMixin'):
    class YamlConfig(mixins.YamlMixin, Config):
        ...


    class LocalYamlConfig(mixins.YamlMixin, LocalConfig):
        ...
else:
    print((
        'PyYAML is not installed. '
        'Run pip install xdgconfig[yaml] to install it.'
    ))

if hasattr(mixins, 'TomlMixin'):
    class TomlConfig(mixins.TomlMixin, Config):
        ...


    class LocalTomlConfig(mixins.TomlMixin, LocalConfig):
        ...
else:
    print((
        'TOML is not installed. '
        'Run pip install xdgconfig[toml] to install it.'
    ))



import unittest

from tests.utils import TestCase
from tests.mocks import MockedJsonConfig


class DefaultConfig:
    _DEFAULTS = {
        'logger.level': 'info',
        'logger.verbosity': 3,
        'app.path': '/path/to/app',
        'app.credentials.username': 'user',
        'app.credentials.password': 'password',
    }


class MockedConfig(DefaultConfig, MockedJsonConfig):
    ...


class TestDefaultConfig(TestCase):
    CONFIG_NAME = 'config.json'

    def test_default_values(self):
        config = self.make_config(MockedConfig, name='defaults.json')
        self.assertEqual(config['logger']['level'], 'info')
        self.assertEqual(config['logger']['verbosity'], 3)
        self.assertEqual(config['app']['path'], '/path/to/app')
        self.assertEqual(config['app']['credentials']['username'], 'user')
        self.assertEqual(config['app']['credentials']['password'], 'password')


if __name__ == '__main__':
    unittest.main()



from typer import Typer
from typer.testing import CliRunner

from xdgconfig.cli_tools import typer_config, make_ctx_settings

from tests.utils import TestCase
from tests.mocks import MockedJsonConfig


class TestArgparseConfig(TestCase):
    CONFIG_NAME = 'config.json'

    def test_config_parsing(self):
        config = self.make_config(MockedJsonConfig, 'argparse')
        runner = CliRunner()

        app = Typer()
        app.add_typer(
            typer_config, name='config',
            context_settings=make_ctx_settings(config)
        )
        result = runner.invoke(app, ['config', '--global', 'key', 'value'])
        print(result.stdout)
        self.assertEqual(result.exit_code, 0)
        self.assertDictEqual(config, {'key': 'value'})



from typing import Any


class defaultdict(dict):
    _DEFAULTS = {}

    def __init__(self, *args, **kwargs):
        parent = kwargs.pop('_parent', None)
        defaults = kwargs.pop('_defaults', {})
        super().__init__(*args, **kwargs)
        self._parent = parent
        self._DEFAULTS = defaults

    def __getitem__(self, key: str) -> Any:
        if key not in self and '.' not in key:
            default = self._default(key)
            if isinstance(default, dict):
                if self._parent is not None:
                    parent = f'{self._parent}.{key}'
                else:
                    parent = key
                default = defaultdict(
                    default, _parent=parent,
                    _defaults=self._DEFAULTS,
                )
            self[key] = default
        if '.' in key:
            return self['.'.join(key.split('.')[1:])]
        return super().__getitem__(key)

    def __getattr__(self, key: str) -> Any:
        if key in dir(self) or key in self.__dict__:
            return self.__getattribute__(key)
        if key in self:
            return self[key]
        k = key.replace(' ', '_').replace("'", '')
        if k in self:
            return self[k]
        raise AttributeError(
            f'Attribute `{key}` does not exist on class `{type(self).__name__}`'
        )

    def __setattr__(self, key:str, value: Any) -> None:
        if key in self:
            self[key] = value
        k = key.replace(' ', '_').replace("'", '')
        if k in self:
            self[k] = value
        super().__setattr__(key, value)

    def _default(self, key: str) -> Any:
        if self._parent:
            path = f'{self._parent}.{key}'
        else:
            path = key
        return self._DEFAULTS.get(
            path, defaultdict(
                _parent=path, _defaults=self._DEFAULTS,
            )
        )



import contextlib

from xdgconfig.serializers import ini, json


__all__ = [
    'JsonMixin', 'IniMixin',
]


class JsonMixin:
    _SERIALIZER = json


class IniMixin:
    _SERIALIZER = ini


with contextlib.suppress(ImportError):
    from xdgconfig.serializers import xml
    __all__.append('XmlMixin')

    class XmlMixin:
        _SERIALIZER = xml


with contextlib.suppress(ImportError):
    from xdgconfig.serializers import toml
    __all__.append('TomlMixin')

    class TomlMixin:
        _SERIALIZER = toml


with contextlib.suppress(ImportError):
    from xdgconfig.serializers import yaml
    __all__.append('YamlMixin')

    class YamlMixin:
        _SERIALIZER = yaml



from xdgconfig import (
    YamlConfig, JsonConfig, XmlConfig,
    TomlConfig, IniConfig,
)

from tests.utils import MockMixin


class MockedYamlConfig(MockMixin, YamlConfig):
    ...


class MockedXmlConfig(MockMixin, XmlConfig):
    ...


class MockedJsonConfig(MockMixin, JsonConfig):
    ...


class MockedTomlConfig(MockMixin, TomlConfig):
    ...


class MockedIniConfig(MockMixin, IniConfig):
    ...



from argparse import ArgumentParser

from xdgconfig.cli_tools import argparse_config

from tests.utils import TestCase
from tests.mocks import MockedJsonConfig


class TestArgparseConfig(TestCase):
    CONFIG_NAME = 'config.json'

    def test_config_parsing(self):
        config = self.make_config(MockedJsonConfig, 'argparse')

        parser = ArgumentParser()
        sp = parser.add_subparsers()
        sp = argparse_config(sp, config)
        args = vars(parser.parse_args(['config', '--global', 'key', 'value']))
        self.assertEqual(args['_global'], True)
        self.assertEqual(args['config_key'], 'key')
        self.assertEqual(args['config_value'], 'value')

        args.pop('callback')(**args)
        self.assertDictEqual(config, {'key': 'value'})



import pathlib

from tests.utils import TestCase
from tests.mocks import MockedIniConfig


class TestIniConfig(TestCase):
    CONFIG_NAME = 'config.ini'

    def test_config_saved(self):
        '''
        Tests that the config file is created correctly.
        '''
        config = self.make_config(MockedIniConfig, 'saved')
        config['string'] = 'string'
        config['integer'] = 0
        config['float'] = 0.1
        config['dict'] = {}
        config['list'] = []
        self.assertEqual(
            config,
            {
                'string': 'string',
                'integer': 0,
                'float': 0.1,
                'dict': {},
                'list': []
            }
        )
        config.save()
        self.assertFileExists(
            (
                pathlib.Path('./__tmp__') /
                self.__class__.__name__ /
                f'saved_{self.CONFIG_NAME}'
            )
        )

    def test_config_loaded(self):
        config = self.make_config(MockedIniConfig, 'load')
        config['string'] = 'string'
        config['integer'] = 0
        config['float'] = 0.1
        config['dict'] = {}
        config['list'] = []
        config.save()

        conf = self.make_config(MockedIniConfig, 'load')
        self.assertEqual(
            conf,
            {
                'string': 'string',
                'integer': 0,
                'float': 0.1,
                'dict': {},
                'list': []
            }
        )

    def test_mutating_subkey(self):
        '''
        Test that mutating a non-existing subkey generates the proper
        tree-like structure.
        '''
        config = self.make_config(MockedIniConfig, 'mutating')
        config['foo']['bar'] = 'baz'
        self.assertEqual(
            config, {'foo': {'bar': 'baz'}}
        )

    def test_config_identity(self):
        '''
        Test that the config file's instance is cached in a Singleton-type
        design pattern. Ensures limited memory footprint when working with
        config files.
        '''
        config = self.make_config(MockedIniConfig, 'identity')
        self.assertIs(config, self.make_config(MockedIniConfig, 'identity'))
        self.assertIsNot(
            config, self.make_config(MockedIniConfig, 'identity_false')
        )



import itertools

from setuptools import setup, find_packages
from xdgconfig import __version__


long_desc = ''
requirements = ['mergedeep>=1.3.4,<2']
with open('readme.md') as f:
    long_desc = f.read()


extras = {
    'jsonc': ['commentjson>=0.9.0,<1'],
    'toml': ['toml>=0.10.2,<1'],
    'yaml': ['PyYAML>=5.4.1,<6'],
    'xml': ['xmltodict>=0.12.0,<1'],
}

for k in range(2, len(extras) + 1):
    for perm in itertools.permutations(
        [x for x in extras if '+' not in x], k
    ):
        extras['+'.join(perm)] = [
            item
            for sublist in [extras[p] for p in perm]
            for item in set(sublist)
        ]

extras['all'] = [
    item
    for ex, sublist in extras.items()
    if '+' not in ex
    for item in set(sublist)
]

setup(
    name='xdgconfig',
    version=__version__,
    author='Dogeek',
    author_email='dogeek@users-noreply.github.com',
    url='https://github.com/dogeek/xdgconfig',
    description='Easy access to `~/.config`',
    long_description=long_desc,
    long_description_content_type='text/markdown',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development :: Libraries',
    ],
    packages=find_packages(),
    install_requires=requirements,
    license='MIT',
    zip_safe=True,
    platforms='any',
    python_requires='>=3.8',
    download_url='https://github.com/dogeek/xdgconfig/releases',
    extras_require=extras,
    keywords=['configuration', 'python3'],
)
